import calendar
import enum
import datetime
import typing

import boto3


class BadDateError(Exception):
    pass


class BadForecastError(Exception):
    pass


PPATableRows = typing.List[typing.List[str]]


class PPATable(typing.TypedDict):
    header: typing.List[str]
    rows: typing.Union[PPATableRows, typing.Dict[str, PPATableRows]]


class PeriodGranularity(enum.Enum):
    DAILY = "DAILY"
    MONTHLY = "MONTHLY"


class CostMetric(enum.Enum):
    AMORTIZED_COST = "AmortizedCost"
    BLENDED_COST = "BlendedCost"
    NET_AMORTIZED_COST = "NetAmortizedCost"
    UNBLENDED_COST = "UnblendedCost"
    NET_UNBLENDED_COST = "NetUnblendedCost"


class CostForecastMetric(enum.Enum):
    AMORTIZED_COST = "AMORTIZED_COST"
    BLENDED_COST = "BLENDED_COST"
    NET_AMORTIZED_COST = "NET_AMORTIZED_COST"
    UNBLENDED_COST = "UNBLENDED_COST"
    NET_UNBLENDED_COST = "NET_UNBLENDED_COST"


class GroupByKey(enum.Enum):
    AZ = "AZ"
    INSTANCE_TYPE = "INSTANCE_TYPE"
    LEGAL_ENTITY = "LEGAL_ENTITY"
    LINKED_ACCOUNT = "LINKED_ACCOUNT"
    OPERATION = "OPERATION"
    PLATFORM = "PLATFORM"
    PURCHASE_TYPE = "PURCHASE_TYPE"
    SERVICE = "SERVICE"
    TAGS = "TAGS"
    TENANCY = "TENANCY"
    RECORD_TYPE = "RECORD_TYPE"
    USAGE_TYPE = "USAGE_TYPE"


INPUT_DATETIME_FORMAT = "%Y-%m-%d"
OUTPUT_DATETIME_FORMAT_MAP = {
    PeriodGranularity.DAILY.value: "%d %B",
    PeriodGranularity.MONTHLY.value: "%B %Y",
}
TODAY = datetime.datetime.now()


def _client(access_key_id: str, secret_access_key: str) -> boto3.client:
    return boto3.client("ce", aws_access_key_id=access_key_id, aws_secret_access_key=secret_access_key)


def _next_day(start: datetime.datetime = TODAY) -> datetime.datetime:
    return start + datetime.timedelta(days=1)


def _next_end_of_the_month(start: datetime.datetime = TODAY) -> datetime.datetime:
    next_day = _next_day(start)
    return datetime.datetime(year=next_day.year, month=next_day.month, day=calendar.monthrange(next_day.year, next_day.month)[1])


def _strf_aws_time_period(time_period: dict, granularity: PeriodGranularity, key: str = "Start") -> str:
    return datetime.datetime.strptime(time_period[key], INPUT_DATETIME_FORMAT).strftime(OUTPUT_DATETIME_FORMAT_MAP[granularity])


def cost(
    access_key_id: str,
    secret_access_key: str,
    start: str,
    end: str = TODAY.strftime(INPUT_DATETIME_FORMAT),
    granularity: PeriodGranularity = PeriodGranularity.MONTHLY.value,
    metric: CostMetric = CostMetric.UNBLENDED_COST.value,
    group_by_key: GroupByKey = GroupByKey.SERVICE.value,
    view_breakdown: bool = True,
) -> PPATable:
    try:
        datetime.datetime.strptime(start, INPUT_DATETIME_FORMAT)
    except Exception as e:
        raise BadDateError(
            f"Parameter 'start' might not have the right format. "
            f"It must be a numeric date in the form year-month-day, like 2020-03-15. Underlying error: {e}"
        ) from None
    raw_data = _client(access_key_id, secret_access_key).get_cost_and_usage(
        TimePeriod={"Start": start, "End": end},
        Granularity=granularity,
        Metrics=[metric],
        GroupBy=[{"Type": "DIMENSION", "Key": group_by_key}],
    )
    if view_breakdown:
        header = [group_by_key, "Cost"]
        rows = {}
    else:
        header = []
        rows = [[]]
    for period in raw_data["ResultsByTime"]:
        title = _strf_aws_time_period(period["TimePeriod"], granularity)
        if view_breakdown:
            rows[title] = sorted(
                [[item["Keys"][0], f"{float(item['Metrics'][metric]['Amount']):.2f}"] for item in period["Groups"]],
                key=lambda item: float(item[1]),
                reverse=True,
            )
        else:
            header.append(title)
            rows[0].append(f"{sum((float(item['Metrics'][metric]['Amount']) for item in period['Groups'])):.2f}")
    return {"header": header, "rows": rows}


def cost_forecast(
    access_key_id: str,
    secret_access_key: str,
    start: str = _next_day().strftime(INPUT_DATETIME_FORMAT),
    end: str = _next_end_of_the_month(_next_day()).strftime(INPUT_DATETIME_FORMAT),
    granularity: PeriodGranularity = PeriodGranularity.MONTHLY.value,
    metric: CostMetric = CostForecastMetric.UNBLENDED_COST.value,
) -> PPATable:
    raw_data = _client(access_key_id, secret_access_key).get_cost_forecast(
        TimePeriod={"Start": start, "End": end}, Granularity=granularity, Metric=metric,
    )
    header = []
    rows = [[]]
    for period in raw_data["ForecastResultsByTime"]:
        header.append(_strf_aws_time_period(period["TimePeriod"], granularity))
        rows[0].append(f"{float(period['MeanValue']):.2f}")
    return {"header": header, "rows": rows}
