"""EC2 - Instance Types

This module contains actions for viewing EC2 instance types.
"""
import uuid
import logging
import functools

from typing import List, Optional, Union

from ppa import ui

from ._ec2 import create_records, paging_request
from .records import EC2InstanceType, EC2InstanceTypeData
from ..records import AWSClient
from .._api import client


TYPE_TABLE_HEADER = ["Name", "Clock Speed (GHz)", "Memory (MiB)", "Hypervisor", "Bare Metal"]
TYPE_TABLE_FIELDS = [
    lambda instance_type: instance_type.get("name"),
    lambda instance_type: instance_type["processor_info"].get("sustained_clock_speed", ""),
    lambda instance_type: instance_type["memory"],
    lambda instance_type: instance_type["hypervisor"],
    lambda instance_type: str(instance_type.get("bare_metal")).title(),
]


logger = logging.getLogger(__name__)

ec2_client = functools.partial(client, endpoint="ec2")
creator = functools.partial(create_records, EC2InstanceTypeData)
data_key = "InstanceTypes"


def get_all(aws_client: AWSClient, region_name: str) -> List[EC2InstanceType]:
    """Get the instance types supported by the supplied EC2 region.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

??? success "Output"

    A list of [EC2InstanceType](../inputs_outputs/#ec2instancetype) dictionaries.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.supported_types:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: instance_types
    ```
    """
    logger.info(f"getting all supported instance types for region {region_name}")
    return creator(paging_request(ec2_client(aws_client=aws_client, region_name=region_name).describe_instance_types, data_key))


def output_table(text: str, types: Union[EC2InstanceType, List[EC2InstanceType]]) -> None:
    """Display a list of instance types in a table.

The table will have the following columns:

- Name
- Clock Speed (GHz)
- Memory (MiB)
- Hypervisor
- Bare Metal

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __types:__ a single or list of [EC2InstanceType](../inputs_outputs/#ec2instancetype) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1" hl_lines="6 11"
    - aws.ec2.instance_types.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: instance_types

    - aws.ec2.instance_types.output_table:
        text: Supported Instance Types
      load:
        types: instance_types
    ```
    """
    if not isinstance(types, list):
        types = [types]
    ui.output_table(text, header=TYPE_TABLE_HEADER, data=[[get(t) for get in TYPE_TABLE_FIELDS] for t in types])


def input_table(
    text: str, types: List[EC2InstanceType], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[EC2InstanceType]:
    """Display a list of instance types in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Clock Speed (GHz)
- Memory (MiB)
- Hypervisor
- Bare Metal

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __types:__ a list of [EC2InstanceType](../inputs_outputs/#ec2instancetype) dictionaries to display in the table

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [EC2InstanceType](../inputs_outputs/#ec2instancetype) dictionaries.

??? tip

    - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

    - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

??? example

    ```YAML linenums="1" hl_lines="5 12"
    - aws.ec2.instance_types.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: instance_types

    - aws.ec2.instance_types.input_table:
        text: Select an Instance Type
        minimum: 1
        maximum: 1
      load:
        types: instance_types
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())

    # Sort by memory otherwise it's a really random scattering.
    sorted_types = sorted(types, key=lambda x: x["memory"])

    while True:
        indices = ui.input_table(
            text,
            header=TYPE_TABLE_HEADER,
            data=[[get(t) for get in TYPE_TABLE_FIELDS] for t in sorted_types],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [sorted_types[i] for i in indices]
