"""EC2 - Instances

This module contains actions for viewing & managing EC2 instances.
"""
import botocore
import uuid
import logging
import functools

from typing import List, Optional, Union, Dict, Any

from ppa import ui

from . import exceptions
from ._ec2 import create_records, nested_paging_request
from .records import EC2Instance, EC2InstanceData, get_tag_value
from ..records import AWSClient
from .._api import client, resource


INSTANCE_TABLE_HEADER = ["Name", "ID", "State", "Public DNS Name", "Private DNS Name", "VPC ID"]
INSTANCE_TABLE_FIELDS = [
    lambda instance: get_tag_value(instance, "Name"),
    lambda instance: instance.get("instance_id"),
    lambda instance: instance.get("state"),
    lambda instance: instance.get("public_dns_name") or "",
    lambda instance: instance.get("private_dns_name") or "",
    lambda instance: instance.get("vpc_id"),
]


logger = logging.getLogger(__name__)

ec2_client = functools.partial(client, endpoint="ec2")
ec2_resource = functools.partial(resource, endpoint="ec2")
creator = functools.partial(create_records, EC2InstanceData)
parent_key = "Reservations"
data_key = "Instances"


def get_by_tags(aws_client: AWSClient, region_name: str, tags: Dict[str, Any], vpc_id: Optional[str] = None) -> List[EC2Instance]:
    """Get instances with _all_ the supplied tag values.

!!! question "Minimum Plugin Version: 2.3.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __tags:__ a dictionary of tag names & values to use in the search

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"

    A list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries.

!!! warning "Case Sensitivity"

    Tags are case sensitive & must match the names & values configured in EC2.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.get_by_tags:
        region_name: eu-west-2
        tags:
          environment: development
          team: engineering
      load:
        aws_client: aws_secrets
      save: development_instances
    ```
    """
    logger.info(f"getting instances with the following tags {tags}")
    filters = [{"Name": f"tag:{tag}", "Values": [value]} for tag, value in tags.items()]
    if vpc_id:
        filters.append({"Name": "vpc-id", "Values": [vpc_id]})
    return creator(
        nested_paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_instances, parent_key, data_key, Filters=filters,
        )
    )


def get_all(aws_client: AWSClient, region_name: str, vpc_id: Optional[str] = None) -> List[EC2Instance]:
    """Get all instances.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"

    A list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.instances.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: all_instances
    ```
    """
    logger.info(f"getting all instances from region {region_name}")
    return creator(
        nested_paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_instances,
            parent_key,
            data_key,
            Filters=[{"Name": "vpc-id", "Values": [vpc_id]}] if vpc_id else [{}],
        )
    )


def get_by_id(aws_client: AWSClient, region_name: str, instance_id: str) -> EC2Instance:
    """Get an instance with a specific instance ID.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __instance_id:__ the instance ID to search for

??? success "Output"

    A single [EC2Instance](../inputs_outputs/#ec2instance) dictionary.

??? example
    ```YAML linenums="1"
    - aws.ec2.instances.get_by_id:
        region_name: eu-west-2
        instance_id: i-0e88060c12d5d7e71
      load:
        aws_client: aws_secrets
      save: instance
    ```
    """
    logger.info(f"getting instance with ID {instance_id} from region {region_name}")
    try:
        return creator(
            nested_paging_request(
                ec2_client(aws_client=aws_client, region_name=region_name).describe_instances,
                parent_key,
                data_key,
                InstanceIds=[instance_id],
            )
        )[0]
    except botocore.exceptions.ClientError as e:
        if "invalid id" in str(e).lower():
            message = f"No instance was found with ID '{instance_id}'."
            ui.output_error(message)
            raise exceptions.ObjectNotFound(message)
        raise e


def get_by_name(aws_client: AWSClient, region_name: str, name: str, vpc_id: Optional[str] = None) -> List[EC2Instance]:
    """Get instances with a specific name.

!!! question "Minimum Plugin Version: 2.0.0"

??? tip "Instance Naming"

    Instances don't have a `name` attribute, so this action searches for the `Name` __tag__ instead.

    This tag is set automatically when you name an instance in the AWS console.

    Unlike security groups, multiple instances can exist with the same name.

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __name:__ the instance name to search for

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"

    A list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.get_by_name:
        region_name: eu-west-2
        name: Ubuntu Server
      load:
        aws_client: aws_secrets
      save: ubuntu_server_instances
    ```
    """
    logger.info(f"getting instance with name {name} from region {region_name}")
    filters = [{"Name": "tag:Name", "Values": [name]}]
    if vpc_id:
        filters.append({"Name": "vpc-id", "Values": [vpc_id]})
    return creator(
        nested_paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_instances, parent_key, data_key, Filters=filters,
        )
    )


def get_stopped(aws_client: AWSClient, region_name: str, vpc_id: Optional[str] = None) -> List[EC2Instance]:
    """Get stopped instances.

This includes all hibernating instances.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"

    A list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.instances.get_stopped:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: stopped_instances
    ```
    """
    logger.info(f"getting stopped instances from region {region_name}")
    filters = [{"Name": "instance-state-name", "Values": ["stopped"]}]
    if vpc_id:
        filters.append({"Name": "vpc-id", "Values": [vpc_id]})
    return creator(
        nested_paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_instances, parent_key, data_key, Filters=filters,
        )
    )


def get_running(aws_client: AWSClient, region_name: str, vpc_id: Optional[str] = None) -> List[EC2Instance]:
    """Get running instances.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"

    A list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.instances.get_running:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: running_instances
    ```
    """
    logger.info(f"getting running instances from region {region_name}")
    filters = [{"Name": "instance-state-name", "Values": ["running"]}]
    if vpc_id:
        filters.append({"Name": "vpc-id", "Values": [vpc_id]})
    return creator(
        nested_paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_instances, parent_key, data_key, Filters=filters,
        )
    )


def get_untagged(aws_client: AWSClient, region_name: str, vpc_id: Optional[str] = None) -> List[EC2Instance]:
    """Get untagged instances.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"

    A list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.instances.get_untagged:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: untagged_instances
    ```
    """
    logger.info(f"getting untagged instances from region {region_name}")
    return [instance for instance in get_all(aws_client, region_name, vpc_id=vpc_id) if not instance["tags"]]


def start(aws_client: AWSClient, region_name: str, instance_id: str) -> None:
    """Start an instance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __instance_id:__ the ID of the instance to start

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.start:
        region_name: eu-west-2
        instance_id: i-0e88060c12d5d7e71
      load:
        aws_client: aws_secrets
    ```
    """
    logger.info(f"starting instance with ID {instance_id} in region {region_name}")
    try:
        return ec2_client(aws_client=aws_client, region_name=region_name).start_instances(InstanceIds=[instance_id])
    except botocore.exceptions.ClientError as e:
        if "invalid id" in str(e).lower():
            raise exceptions.ObjectNotFound(f"No instance found with ID '{instance_id}'.")
        raise e


def launch(
    aws_client: AWSClient,
    region_name: str,
    image_id: str,
    name: Optional[str] = None,
    instance_type: Optional[str] = None,
    key_name: Optional[str] = None,
    subnet_id: Optional[str] = None,
) -> EC2Instance:
    """Launch a new instance from an AMI.


!!! question "Minimum Plugin Version: 2.2.0"


??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __image_id:__ the ID of the AMI to launch the new instance from

    - __name:__ a name for the new instance (defaults to empty)

    - __instance_type:__ the instance type (defaults to __t2.micro__, see [EC2 Instance Types](https://aws.amazon.com/ec2/instance-types/) for more information)

    - __key_name:__ the name of the key-pair to use (defaults to empty)

    - __subnet_id:__ the ID of the subnet to launch the instance in (defaults to the default VPC subnet)


??? success "Output"

    A single [EC2Instance](../inputs_outputs/#ec2instance) dictionary containing the new instance.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.launch:
        region_name: eu-west-2
        image_id: ami-0a669382ea0feb73a
        instance_type: t2.micro
        key_name: linux-key-pair
        name: new-instance
      load:
        aws_client: aws_secrets
      save: new_instance
    ```
    """
    logger.info(f"launching new instance from AMI {image_id} in region {region_name}")
    kwargs = {"ImageId": image_id, "MinCount": 1, "MaxCount": 1, "InstanceType": instance_type or "t2.micro"}

    # If the following are None, boto3 won't fall back to their default
    # values. It'll just pass None through as a value which raises an exception.
    if key_name is not None:
        kwargs["KeyName"] = key_name
    if subnet_id is not None:
        kwargs["SubnetId"] = subnet_id
    if name is not None:
        # If a TagSpecifications list is supplied, it has to include the ResourceType.
        kwargs["TagSpecifications"] = [{"ResourceType": "instance", "Tags": [{"Key": "Name", "Value": name}]}]

    return get_by_id(
        aws_client, region_name, ec2_resource(aws_client=aws_client, region_name=region_name).create_instances(**kwargs)[0].id
    )


def stop(aws_client: AWSClient, region_name: str, instance_id: str, hibernate: Optional[bool] = False) -> None:
    """Stop an instance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __instance_id:__ the ID of the instance to stop

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.stop:
        region_name: eu-west-2
        instance_id: i-0e88060c12d5d7e71
      load:
        aws_client: aws_secrets
    ```
    """
    logger.info(f"stopping instance with ID {instance_id} in region {region_name}")
    try:
        return ec2_client(aws_client=aws_client, region_name=region_name).stop_instances(
            InstanceIds=[instance_id], Hibernate=hibernate
        )
    except botocore.exceptions.ClientError as e:
        if "invalid id" in str(e).lower():
            raise exceptions.ObjectNotFound(f"No instance found with ID '{instance_id}'.")
        raise e


def terminate(aws_client: AWSClient, region_name: str, instance_id: str) -> None:
    """Terminate an instance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __instance_id:__ the ID of the instance to terminate

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.terminate:
        region_name: eu-west-2
        instance_id: i-0e88060c12d5d7e71
      load:
        aws_client: aws_secrets
    ```
    """
    logger.info(f"terminating instance with ID {instance_id} in region {region_name}")
    try:
        return ec2_client(aws_client=aws_client, region_name=region_name).terminate_instances(InstanceIds=[instance_id])
    except botocore.exceptions.ClientError as e:
        if "invalid id" in str(e).lower():
            raise exceptions.ObjectNotFound(f"No instance found with ID '{instance_id}'.")
        raise e


def add_to_security_group(aws_client: AWSClient, region_name: str, instance_id: str, security_group_id: str) -> None:
    """Add an instance to a security group.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __instance_id:__ the instance ID

    - __security_group_id:__ the security group ID

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.add_to_security_group:
        region_name: eu-west-2
      load:
        instance_id: instance.id
        security_group_id: security_group.id
        aws_client: aws_secrets
    ```
    """
    logger.info(f"adding instance with ID {instance_id} in region {region_name} to security group with ID {security_group_id}")
    instance = get_by_id(aws_client, region_name, instance_id)
    group_ids = [group["group_id"] for group in instance["security_groups"]]

    if security_group_id in group_ids:
        # Nothing to be done here.
        logger.info(
            f"instance with ID {instance_id} in region {region_name} is already in security group with ID {security_group_id}"
        )
        return

    group_ids.append(security_group_id)
    try:
        next(
            iter(ec2_resource(aws_client=aws_client, region_name=region_name).instances.filter(InstanceIds=[instance_id]))
        ).modify_attribute(Groups=group_ids)
    except botocore.exceptions.ClientError as e:

        def _raise_with_message(exception, message):
            logger.error(message)
            ui.output_error(message)
            raise exception(message)

        ec2_error = str(e).lower()
        base_message = f"Cannot add instance to security group '{security_group_id}'"
        if "that belong to different networks" in ec2_error:
            _raise_with_message(ValueError, f"{base_message} as they are on different networks.")
        elif "maximum number of security groups" in ec2_error:
            _raise_with_message(
                RuntimeError, f"{base_message}. The maximum number of security groups per-instance has been reached."
            )
        raise e


def remove_from_security_group(aws_client: AWSClient, region_name: str, instance_id: str, security_group_id: str) -> None:
    """Remove an instance from a security group.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __instance_id:__ the instance ID

    - __security_group_id:__ the security group ID

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.remove_from_security_group:
        region_name: eu-west-2
      load:
        instance_id: instance.id
        security_group_id: security_group.id
        aws_client: aws_secrets
    ```
    """
    logger.info(
        f"removing instance with ID {instance_id} in region {region_name} from security group with ID {security_group_id}"
    )
    instance = get_by_id(aws_client, region_name, instance_id)
    group_ids = [group["group_id"] for group in instance["security_groups"]]

    if security_group_id not in group_ids:
        # Nothing to be done here.
        logger.info(
            f"instance with ID {instance_id} in region {region_name} is not in security group with ID {security_group_id}"
        )
        return

    group_ids.remove(security_group_id)
    next(
        iter(ec2_resource(aws_client=aws_client, region_name=region_name).instances.filter(InstanceIds=[instance_id]))
    ).modify_attribute(Groups=group_ids)


def tag(aws_client: AWSClient, region_name: str, instance_id: str, tags: Dict[str, str]) -> None:
    """Apply one or more tags to an instance.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __instance_id:__ the instance ID

    - __tags:__ a dictionary of tags to apply to the instance

!!! warning "Overwriting Tags"

    If any tags supplied to this action already exist on the instance, their values will be overwritten.

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws.ec2.instances.tag:
        region_name: eu-west-2
        tags:
          environment: uat
          replicated: false
      load:
        instance_id: instance.id
        aws_client: aws_secrets
    ```
    """
    logger.info(f"adding tags {tags} to instance with ID {instance_id} in region {region_name}")

    try:
        formatted_tags = [{"Key": key, "Value": str(value)} for key, value in tags.items()]
    except Exception as e:
        message = "Failed to convert one or more of the supplied tags to a string for EC2."
        logger.exception(e)
        logger.error(message)
        ui.output_error(message)
        raise ValueError(message)

    ec2_client(aws_client=aws_client, region_name=region_name).create_tags(Resources=[instance_id], Tags=formatted_tags)


def output_table(text: str, instances: Union[EC2Instance, List[EC2Instance]]) -> None:
    """Display a list of instances in a table.

The table will have the following columns:

- Name
- ID
- State
- Public DNS Name
- Private DNS Name
- VPC ID

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __instances:__ a single or list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example "Single Instance"

    Finding an instance with [get_by_id](#get_by_id), saving it as `instance`, & displaying it:

    ```YAML linenums="1" hl_lines="6 11"
    - aws.ec2.instances.get_by_id:
        region_name: eu-west-2
        instance_id: i-0e88060c12d5d7e71
      load:
        aws_client: aws_secrets
      save: instance

    - aws.ec2.instances.output_table:
        text: Instance i-0e88060c12d5d7e71
      load:
        instances: instance
    ```

??? example "Multiple Instances"

    Finding all instances with [get_all](#get_all), saving them as `instances`, & displaying them:

    ```YAML linenums="1" hl_lines="5 10"
    - aws.ec2.instances.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: instances

    - aws.ec2.instances.output_table:
        text: All Instances
      load:
        instances: instances
    ```
    """
    if not isinstance(instances, list):
        instances = [instances]

    ui.output_table(
        text, header=INSTANCE_TABLE_HEADER, data=[[get(instance) for get in INSTANCE_TABLE_FIELDS] for instance in instances]
    )


def input_table(
    text: str, instances: List[EC2Instance], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[EC2Instance]:
    """Display a list of instances in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- ID
- State
- Public DNS Name
- Private DNS Name
- VPC ID

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __instances:__ a list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries to display in the table

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [EC2Instance](../inputs_outputs/#ec2instance) dictionaries.

??? tip

    - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

    - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

??? example

    Finding instances with [get_all](#get_all), saving them as `groups`, & waiting for a single selection:

    ```YAML linenums="1" hl_lines="5 12 13"
    - aws.ec2.instances.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: instances

    - aws.ec2.instances.input_table:
        text: Choose an instance
        minimum: 1
        maximum: 1
      load:
        instances: instances
      save: selected_instance
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=INSTANCE_TABLE_HEADER,
            data=[[get(instance) for get in INSTANCE_TABLE_FIELDS] for instance in instances],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [instances[i] for i in indices]
