"""EC2 - Key Pairs

This module contains actions for viewing EC2 Key Pairs.
"""
import logging
import uuid
import functools

from typing import List, Optional, Union

from ppa import ui

from .records import EC2KeyPair, EC2KeyPairData
from ._ec2 import create_records, paging_request
from ..records import AWSClient
from .._api import client


logger = logging.getLogger(__name__)

ec2_client = functools.partial(client, endpoint="ec2")
data_key = "KeyPairs"
creator = functools.partial(create_records, EC2KeyPairData)


KEY_PAIR_TABLE_HEADER = ["Name", "ID"]
KEY_PAIR_TABLE_FIELDS = [
    lambda key_pair: key_pair.get("name"),
    lambda key_pair: key_pair.get("key_pair_id"),
]


def get_all(aws_client: AWSClient, region_name: str) -> List[EC2KeyPair]:
    """Get all key pairs.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

??? success "Output"

    A list of [EC2KeyPair](../inputs_outputs/#ec2keypair) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.key_pairs.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: all_key_pairs
    ```
    """
    logger.info(f"getting all key pairs from region {region_name}")
    return creator(paging_request(ec2_client(aws_client=aws_client, region_name=region_name).describe_key_pairs, data_key))


def output_table(text: str, key_pairs: Union[EC2KeyPair, List[EC2KeyPair]]) -> None:
    """Display a list of key pairs in a table.

The table will have the following columns:

- Name
- ID

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __key_pairs:__ a single or list of [EC2KeyPair](../inputs_outputs/#ec2keypair) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1" hl_lines="6 11"
    - aws.ec2.key_pairs.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: all_key_pairs

    - aws.ec2.key_pairs.output_table:
        text: All Key Pairs
      load:
        key_pairs: all_key_pairs
    ```

    """
    key_pairs = [key_pairs] if not isinstance(key_pairs, list) else key_pairs
    ui.output_table(
        text, header=KEY_PAIR_TABLE_HEADER, data=[[get(key_pair) for get in KEY_PAIR_TABLE_FIELDS] for key_pair in key_pairs]
    )


def input_table(
    text: str, key_pairs: List[EC2KeyPair], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[EC2KeyPair]:
    """Display a list of key pairs in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- ID

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __key_pairs:__ a list of [EC2KeyPair](../inputs_outputs/#ec2keypair) dictionaries to display in the table

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [EC2KeyPair](../inputs_outputs/#ec2keypair) dictionaries.

??? tip

    - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

    - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

??? example

    ```YAML linenums="1"
    - aws.ec2.key_pairs.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: all_key_pairs

    - aws.ec2.key_pairs.input_table:
        text: Select a key pair
        minimum: 1
        maximum: 1
      load:
        key_pairs: all_key_pairs
      save: selected_key_pair
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=KEY_PAIR_TABLE_HEADER,
            data=[[get(key_pair) for get in KEY_PAIR_TABLE_FIELDS] for key_pair in key_pairs],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [key_pairs[i] for i in indices]
