import logging

from typing import TypedDict, List, Union, Optional

from datacaster.classes import CastDataClass

from .._converters import optional_date_string
from ..records import Tag


DateString = Optional[str]


logger = logging.getLogger(__name__)


class CpuOptions(TypedDict):
    core_count: int
    threads_per_core: int


class MinimisedGroup(TypedDict):
    name: str
    group_id: str


class AWSClient(TypedDict):
    """
- Contains AWS API connection details

- It is used as an _input_ to many of the actions in this plugin

!!! tip "Generating API Access Keys"

    See this [AWS article](https://aws.amazon.com/blogs/security/wheres-my-secret-access-key/) if you need to generate API access keys.

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

!!! danger "Required Keys"

    __access_key_id__: AWS access key ID

    __secret_access_key__: AWS secret access key

!!! example

    ```yaml linenums="1"
    aws_client:
      access_key_id: your_access_key_id
      secret_access_key: your_secret_access_key
    ```
    """

    access_key_id: str
    secret_access_key: str


class EC2KeyPairData(CastDataClass):
    key_pair_id: str
    fingerprint: str
    name: str
    tags: List[Tag] = []

    __class_config__ = {
        "rename_fields": {"KeyPairId": "key_pair_id", "KeyFingerprint": "fingerprint", "KeyName": "name", "Tags": "tags"},
        "cast_functions": {
            "fields": {"tags": lambda tags: [{"key": tag_dict["Key"], "value": tag_dict["Value"]} for tag_dict in tags]}
        },
    }


class EC2KeyPair(TypedDict):
    """
- Contains EC2 key pair properties

??? info "EC2KeyPair Keys"

    __name__: name of the key pair

    __key_pair_id__: the ID of the key pair

    __fingerprint__: the key fingerprint

    __tags__: a list of dictionaries containing the `key` & `value` of each tag associated with the key pair

??? example "YAML Structure"

    ```yaml linenums="1"
    name: uat-ubuntu-servers
    key_pair_id: key-08e7e20d238ce6rc5
    fingerprint: 17:ef:57:2d:5b:b4:bb:f2:9f:09:09:fc:d4:f7:3c:23:b4:04:14:f6
    tags:
      - key: environment
        value: UAT
    ```
    """

    __annotations__ = EC2KeyPairData.__annotations__


class EC2IPRangeData(CastDataClass):
    cidr_ip: str
    description: str

    __class_config__ = {"rename_fields": {"CidrIp": "cidr_ip", "Description": "description"}}


class EC2IPRange(TypedDict):
    """
- Contains IP range properties associated with an inbound/outbound rule

??? info "EC2IPRange Keys"

    __cidr_ip__: allowed source/destination CIDR address

    __description__: description for the IP range associated with the rule

??? example "YAML Structure"

    ```yaml linenums="1"
    cidr_ip: 1.2.3.4/5
    description: SSH Access from Corporate Office
    ```
    """

    __annotations__ = EC2IPRangeData.__annotations__


class EC2InboundRuleData(CastDataClass):
    protocol: str
    ip_ranges: List[EC2IPRange] = []
    from_port: int
    to_port: int
    __class_config__ = {
        "rename_fields": {"IpProtocol": "protocol", "IpRanges": "ip_ranges", "FromPort": "from_port", "ToPort": "to_port"},
        "cast_functions": {
            "fields": {"ip_ranges": lambda ip_ranges: [EC2IPRangeData(**ip_range).__dict__ for ip_range in ip_ranges]}
        },
    }


class EC2InboundRule(TypedDict):
    """
- Contains properties for an inbound rule associated with an EC2 security group

??? info "EC2InboundRule Keys"

    __protocol__: rule protocol (`tcp`, `udp`, or `-1` for any)

    __ip_ranges__: list of [EC2IPRanges](#iprange) associated with the rule

    __from_port__: rule start port

    __to_port__: rule end port

??? example "YAML Structure"

    ```yaml linenums="1"
    protocol: tcp
    ip_ranges:
      - cidr_ip: 1.2.3.4/5
        description: SSH Access from Corporate Office
    from_port: 22
    to_port: 22
    ```
    """

    __annotations__ = EC2InboundRuleData.__dict__["__annotations__"]


class EC2OutboundRuleData(CastDataClass):
    protocol: str
    ip_ranges: List[EC2IPRange] = []
    from_port: int
    to_port: int
    __class_config__ = {
        "rename_fields": {"IpProtocol": "protocol", "IpRanges": "ip_ranges", "FromPort": "from_port", "ToPort": "to_port"},
        "cast_functions": {
            "fields": {"ip_ranges": lambda ip_ranges: [EC2IPRangeData(**ip_range).__dict__ for ip_range in ip_ranges]}
        },
    }


class EC2OutboundRule(TypedDict):
    """
- Contains properties for an outbound rule associated with an EC2 security group

??? info "EC2OutboundRule Keys"

    __protocol__: rule protocol (`tcp`, `udp`, or `-1` for any)

    __ip_ranges__: list of [EC2IPRanges](#iprange) associated with the rule

    __from_port__: rule start port

    __to_port__: rule end port

??? example "YAML Structure"

    ```yaml linenums="1"
    protocol: tcp
    ip_ranges:
    - cidr_ip: 1.2.3.4/5
      description: Outbound Access to Corporate Office
    from_port: 22
    to_port: 22
    ```
    """

    __annotations__ = EC2OutboundRuleData.__dict__["__annotations__"]


class EC2SecurityGroupData(CastDataClass):
    description: str
    name: str
    owner_id: str
    group_id: str
    vpc_id: str
    inbound_rules: List[EC2InboundRule] = []
    outbound_rules: List[EC2OutboundRule] = []
    tags: List[Tag] = []
    __class_config__ = {
        "rename_fields": {
            "Description": "description",
            "GroupName": "name",
            "IpPermissions": "inbound_rules",
            "IpPermissionsEgress": "outbound_rules",
            "OwnerId": "owner_id",
            "GroupId": "group_id",
            "Tags": "tags",
            "VpcId": "vpc_id",
        },
        "cast_functions": {
            "fields": {
                "inbound_rules": lambda rules: [EC2InboundRuleData(**rule).__dict__ for rule in rules],
                "outbound_rules": lambda rules: [EC2OutboundRuleData(**rule).__dict__ for rule in rules],
                "tags": lambda tags: [{"key": tag_dict["Key"], "value": tag_dict["Value"]} for tag_dict in tags],
            }
        },
    }


class EC2SecurityGroup(TypedDict):
    """
- Contains EC2 security group properties

??? info "EC2SecurityGroup Keys"

    __description__: description of the security group

    __name__: name of the security group

    __owner_id__: ID of the security group owner

    __group_id__: ID of the security group

    __vpc_id__: ID of the security group VPC

    __inbound_rules__: a list of [EC2InboundRules](#ec2inboundrule) configured in the security group

    __outbound_rules__: a list of [EC2OutboundRules](#ec2outboundrule) configured in the security group

    __tags__: a list of dictionaries containing the `key` & `value` of each tag associated with the security group

??? example "YAML Structure"

    ```yaml linenums="1"
    description: SSH access on port 22
    name: SSH
    owner_id: 381957475495
    group_id: 'sg-04cd36d28fe84e079'
    vpc_id: vpc-1c0be4g4
    inbound_rules:
      - protocol: tcp
        ip_ranges:
        - cidr_ip: 1.2.3.4/5
          description: SSH Access from Corporate Office
        from_port: 22
        to_port: 22
    outbound_rules:
      - protocol: tcp
        ip_ranges:
        - cidr_ip: 1.2.3.4/5
          description: Outbound Access to Corporate Office
        from_port: 22
        to_port: 22
    tags:
      - key: environment
        value: production
    ```
    """

    __annotations__ = EC2SecurityGroupData.__annotations__


class EC2VPCData(CastDataClass):
    cidr_block: str
    is_default: bool
    state: str
    vpc_id: str
    tags: List[Tag] = []
    __class_config__ = {
        "rename_fields": {
            "CidrBlock": "cidr_block",
            "IsDefault": "is_default",
            "State": "state",
            "VpcId": "vpc_id",
            "Tags": "tags",
        },
        "cast_functions": {
            "fields": {"tags": lambda tags: [{"key": tag_dict["Key"], "value": tag_dict["Value"]} for tag_dict in tags]},
        },
    }


class EC2VPC(TypedDict):
    """
- Contains properties for an EC2 Virtual Private Cloud

??? info "EC2VPC Keys"

    __cidr_block__: CIDR address range for the VPC

    __is_default__: indicates whether it is the default VPC

    __state__: VPC state

    __vpc_id__: ID of the VPC

    __tags__: a list of dictionaries containing the `key` & `value` of each tag associated with the VPC

??? example "YAML Structure"

    ```yaml linenums="1"
    cidr_block: 172.31.0.0/16
    is_default: true
    state: available
    vpc_id: vpc-0d7fbc6c4aa6c1ee3
    tags:
      - key: name
        value: Default VPC
    ```
    """

    __annotations__ = EC2VPCData.__annotations__


class EC2NetworkInterfaceData(CastDataClass):
    availability_zone: str
    description: str
    groups: List[MinimisedGroup] = []
    interface_type: str
    network_interface_id: str
    private_dns_name: str
    private_ip_address: str
    status: str
    vpc_id: str
    __class_config__ = {
        "rename_fields": {
            "AvailabilityZone": "availability_zone",
            "Description": "description",
            "Groups": "groups",
            "InterfaceType": "interface_type",
            "NetworkInterfaceId": "network_interface_id",
            "PrivateDnsName": "private_dns_name",
            "PrivateIpAddress": "private_ip_address",
            "Status": "status",
            "VpcId": "vpc_id",
        },
        "cast_functions": {
            "fields": {
                "groups": lambda group_list: [{"name": group["GroupName"], "group_id": group["GroupId"]} for group in group_list],
            },
        },
    }


class EC2NetworkInterface(TypedDict):
    __annotations__ = EC2NetworkInterfaceData.__annotations__


class EC2InstanceData(CastDataClass):
    image_id: str
    instance_id: str
    instance_type: str
    key_name: str
    launch_time: Optional[DateString]
    monitoring: str  # state key
    platform: str
    private_dns_name: str
    private_ip_address: str
    public_dns_name: str
    state: str  # state key
    vpc_id: str
    subnet_id: str
    security_groups: List[MinimisedGroup] = []
    root_device_type: str
    tags: List[Tag] = []
    cpu_options: CpuOptions
    __class_config__ = {
        "rename_fields": {
            "ImageId": "image_id",
            "InstanceId": "instance_id",
            "InstanceType": "instance_type",
            "AmiLaunchIndex": "ami_launch_index",
            "KeyName": "key_name",
            "LaunchTime": "launch_time",
            "Monitoring": "monitoring",
            "PrivateDnsName": "private_dns_name",
            "PrivateIpAddress": "private_ip_address",
            "PublicDnsName": "public_dns_name",
            "State": "state",
            "VpcId": "vpc_id",
            "SubnetId": "subnet_id",
            "SecurityGroups": "security_groups",
            "RootDeviceType": "root_device_type",
            "Tags": "tags",
            "CpuOptions": "cpu_options",
            "Platform": "platform",
        },
        "cast_functions": {
            "fields": {
                "monitoring": lambda data: data["State"],
                "state": lambda data: data["Name"],
                "security_groups": lambda group_list: [
                    {"name": group["GroupName"], "group_id": group["GroupId"]} for group in group_list
                ],
                "tags": lambda tags: [{"key": tag_dict["Key"], "value": tag_dict["Value"]} for tag_dict in tags],
                "cpu_options": lambda options: {
                    "core_count": options["CoreCount"],
                    "threads_per_core": options["ThreadsPerCore"],
                },
            },
            "types": {Optional[DateString]: optional_date_string},
        },
    }


class EC2Instance(TypedDict):
    """
- Contains EC2 instance properties

??? info "EC2Instance Keys"

    __image_id__: ID of the image used by the instance

    __instance_id__: ID of the instance

    __instance_type__: type of the instance

    __key_name__: name of the public/private key pair associated with the instance

    __launch_time__: time of launch (format: `06:45:23 09/07/2020`)

    __monitoring__: state of instance monitoring (`enabled` or `disabled`)

    __platform__: platform type (`windows` or None)

    __private_dns_name__: instance private dns name

    __private_ip_address__: instance private IP address

    __public_dns_name__: instance public DNS name

    __state__: instance state (`starting`, `running`, `stopping`, `stopped`, or `terminated`)

    __vpc_id__: ID of the instance VPC

    __subnet_id__: ID of the subnet the instance was launched in

    __security_groups__: list of dictionaries containing the `name` & `group_id` of each [EC2SecurityGroup](#securitygroup) associated with the instance

    __root_device_type__: type of the instance root volume (`ebs` or an instance store volume)

    __tags__: a list of dictionaries containing the `key` & `value` of each tag associated with the instance

    __cpu_options__: instance CPU options dictionary containing `core_count` & `threads_per_core`

??? example "YAML Structure"

    ```yaml linenums="1"
    image_id: ami-00f6a0c18edb19300
    instance_id: i-0e88060c12d5d7e71
    instance_type: t2.micro
    key_name: ubuntu-server
    launch_time: 06:45:23 09/07/2020
    monitoring: disabled
    platform: windows
    private_dns_name: ip-172-31-5-236.eu-west-2.compute.internal
    private_ip_address: 172.31.5.236
    public_dns_name:
    state: stopped
    vpc_id: vpc-9e5cd4d5
    subnet_id: subnet-80c3d1e7
    security_groups:
    - name: SSH Access
      group_id: sg-033f2b3f97134d512
    root_device_type: ebs
    tags:
    - key: Name
      value: Ubuntu Server
    - key: Environment
      value: UAT
    cpu_options:
      core_count: 2
      threads_per_core: 4
    ```
    """

    __annotations__ = EC2InstanceData.__annotations__


def get_tag_value(item: Union[EC2NetworkInterface, EC2SecurityGroup, EC2VPC, EC2Instance], tag_name: str) -> Union[str, None]:
    try:
        return next(iter([tag_dict["value"] for tag_dict in item["tags"] if tag_dict["key"] == tag_name]))
    except StopIteration:
        return ""


class EC2InstanceProcessorInfoData(CastDataClass):
    supported_architectures: List[str]
    sustained_clock_speed: float

    __class_config__ = {
        "rename_fields": {
            "SupportedArchitectures": "supported_architectures",
            "SustainedClockSpeedInGhz": "sustained_clock_speed",
        }
    }


class EC2InstanceProcessorInfo(TypedDict):
    """
- Contains EC2 Instance type processor information

??? info "EC2InstanceProcessorInfo Keys"

    __supported_architectures__: a list of processor architectures supported by the instance type

    __sustained_clock_speed__: the sustained clock speed in __GHz__


??? example "YAML Structure"

    ```yaml linenums="1"
    supported_architectures:
      - x86_64
    sustained_clock_speed: 3.1
    ```
    """

    __annotations__ = EC2InstanceProcessorInfoData.__annotations__


class EC2InstanceVCPUInfoData(CastDataClass):
    default_vcpus: int
    default_cores: int
    default_threads_per_core: int
    valid_cores: List[int]
    valid_threads_per_core: List[int]

    __class_config__ = {
        "rename_fields": {
            "DefaultVCpus": "default_vcpus",
            "DefaultCores": "default_cores",
            "DefaultThreadsPerCore": "default_threads_per_core",
            "ValidCores": "valid_cores",
            "ValidThreadsPerCore": "valid_threads_per_core",
        }
    }


class EC2InstanceVCPUInfo(TypedDict):
    """
- Contains EC2 Instance type VCPU information

??? info "EC2InstanceVCPUInfo Keys"

    __default_vcpus__: the default number of VCPUs

    __default_cores__: the default number of VCPU cores

    __default_threads_per_core__: the default number of threads per VCPU core

    __valid_cores__: a list of valid core numbers

    __valid_threads_per_core__: a list of valid threads per core


??? example "YAML Structure"

    ```yaml linenums="1"
    default_vcpus: 2
    default_cores: 1
    default_threads_per_core: 2
    valid_cores:
      - 1
    valid_threads_per_core:
      - 1
      - 2
    ```
    """

    __annotations__ = EC2InstanceVCPUInfoData.__annotations__


class EC2InstanceTypeData(CastDataClass):
    name: str
    bare_metal: bool
    free_tier_eligible: bool
    supported_usage_classes: List[str]
    hypervisor: str
    vcpu_info: EC2InstanceVCPUInfo
    processor_info: EC2InstanceProcessorInfo
    memory: int

    __class_config__ = {
        "rename_fields": {
            "InstanceType": "name",
            "FreeTierEligible": "free_tier_eligible",
            "SupportedUsageClasses": "supported_usage_classes",
            "Hypervisor": "hypervisor",
            "VCpuInfo": "vcpu_info",
            "ProcessorInfo": "processor_info",
            "BareMetal": "bare_metal",
            "MemoryInfo": "memory",
        },
        "cast_functions": {
            "fields": {
                "vcpu_info": lambda vcpu_info: EC2InstanceVCPUInfoData(**vcpu_info).__dict__,
                "processor_info": lambda processor_info: EC2InstanceProcessorInfoData(**processor_info).__dict__,
                "memory": lambda memory_info: memory_info.get("SizeInMiB"),
            }
        },
    }


class EC2InstanceType(TypedDict):
    """
- Contains properties for an EC2 Instance type

??? info "EC2InstanceType Keys"

    __name__: the name of the instance type

    __bare_metal__: indicates whether the instance type is bare metal

    __free_tier_eligible__: indicates whether the instance type is free tier eligible

    __supported_usage_classes__: indicates whether the instance type is offered for __spot__, __on-demand__, or both

    __hypervisor__: the instance type hypervisor (usually __xen__ or __nitro__)

    __memory__: the memory of the instance type in __MiB__

    __vcpu_info__: an [EC2InstanceVCPUInfo](#ec2instancevcpuinfo) dictionary containing instance type VCPU details

    __processor_info__: an [EC2InstanceProcessorInfo](#ec2instanceprocessorinfo) dictionary containing instance type processor details


??? example "YAML Structure"

    ```yaml linenums="1"
    name: t3.nano
    bare_metal: false
    free_tier_eligible: false
    supported_usage_classes:
      - on-demand
      - spot
    hypervisor: nitro
    vcpu_info:
      default_vcpus: 2
      default_cores: 1
      default_threads_per_core: 2
      valid_cores:
        - 1
      valid_threads_per_core:
        - 1
        - 2
    processor_info:
      supported_architectures:
        - x86_64
      sustained_clock_speed: 2.5
    memory: 512
    ```
    """

    __annotations__ = EC2InstanceTypeData.__annotations__
