"""EC2 - Security Groups

This module contains actions for viewing & managing EC2 security groups.
"""
import re
import uuid
import logging
import functools

from typing import List, Optional, Union

import botocore

from ppa import ui

from . import exceptions, instances, network_interfaces
from ._ec2 import create_records, paging_request
from .records import EC2SecurityGroup, EC2SecurityGroupData
from ..records import AWSClient
from .._api import client


GROUP_TABLE_HEADER = ["Name", "ID", "Description", "Inbound IPv4 Rule Count", "Outbound IPv6 Rule Count", "VPC ID"]
GROUP_TABLE_FIELDS = [
    lambda group: group.get("name"),
    lambda group: group.get("group_id"),
    lambda group: group.get("description") or "",
    lambda group: _count_rules(group, "inbound_rules"),
    lambda group: _count_rules(group, "outbound_rules"),
    lambda group: group.get("vpc_id"),
]


logger = logging.getLogger(__name__)

ec2_client = functools.partial(client, endpoint="ec2")
creator = functools.partial(create_records, EC2SecurityGroupData)
data_key = "SecurityGroups"


def _count_rules(group: EC2SecurityGroup, key: str) -> int:
    return len(list(sum((rule.get("ip_ranges") or [] for rule in group.get(key) or []), [],)))


def get_all(aws_client: AWSClient, region_name: str, vpc_id: Optional[str] = None) -> List[EC2SecurityGroup]:
    """Get all security groups.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"
    A list of [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: all_security_groups
    ```
    """
    return creator(
        paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_security_groups,
            data_key,
            Filters=[{"Name": "vpc-id", "Values": [vpc_id]}] if vpc_id else [{}],
        )
    )


def get_by_name(aws_client: AWSClient, region_name: str, name: str, vpc_id: Optional[str] = None) -> EC2SecurityGroup:
    """Get a security group with a specific name.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __name:__ the security group name to search for

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"
    A single [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionary.

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.get_by_name:
        region_name: eu-west-2
        name: Ubuntu SSH Access
      load:
        aws_client: aws_secrets
      save: ubuntu_ssh_group
    ```
    """
    filters = [{"Name": "group-name", "Values": [name]}]
    if vpc_id:
        filters.append({"Name": "vpc-id", "Values": [vpc_id]})
    try:
        return creator(
            paging_request(
                ec2_client(aws_client=aws_client, region_name=region_name).describe_security_groups, data_key, Filters=filters
            )
        )[0]
    except botocore.exceptions.ClientError as e:
        if "does not exist" in str(e).lower():
            raise exceptions.ObjectNotFound(f"No security group found with name '{name}'.")
        raise e


def get_by_id(aws_client: AWSClient, region_name: str, group_id: str) -> EC2SecurityGroup:
    """Get a security group with a specific group ID.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __group_id:__ the security group ID to search for

??? success "Output"
    A single [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionary.

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.get_by_id:
        region_name: eu-west-2
        group_id: sg-04cd36d28fe84e079
      load:
        aws_client: aws_secrets
      save: security_group
    ```
    """
    try:
        return creator(
            paging_request(
                ec2_client(aws_client=aws_client, region_name=region_name).describe_security_groups,
                data_key,
                Filters=[{"Name": "group-id", "Values": [group_id]}],
            )
        )[0]
    except botocore.exceptions.ClientError as e:
        if "invalid id" in str(e).lower():
            raise exceptions.ObjectNotFound(f"No security group found with ID '{group_id}'.")
        raise e


def get_by_inbound_port(
    aws_client: AWSClient, region_name: str, port: int, vpc_id: Optional[str] = None
) -> List[EC2SecurityGroup]:
    """Get all security groups that have inbound rules for the supplied port.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __port:__ the inbound port

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"
    A list of [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.get_by_inbound_port:
        region_name: eu-west-2
        port: 22
      load:
        aws_client: aws_secrets
      save: ssh_enabled_groups
    ```
    """
    filters = [{"Name": "ip-permission.to-port", "Values": [str(port)]}]
    if vpc_id:
        filters.append({"Name": "vpc-id", "Values": [vpc_id]})
    return creator(
        paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_security_groups, data_key, Filters=filters
        )
    )


def get_by_inbound_cidr(
    aws_client: AWSClient, region_name: str, cidr: str, vpc_id: Optional[str] = None
) -> List[EC2SecurityGroup]:
    """Get all security groups that have inbound rules for the supplied CIDR address.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __cidr:__ the source CIDR address

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"
    A list of [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.get_by_inbound_cidr:
        region_name: eu-west-2
        cidr: 1.2.3.4/5
      load:
        aws_client: aws_secrets
      save: public_access_groups
    ```
    """
    filters = [{"Name": "ip-permission.cidr", "Values": [cidr]}]
    if vpc_id:
        filters.append({"Name": "vpc-id", "Values": [vpc_id]})
    return creator(
        paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_security_groups, data_key, Filters=filters
        )
    )


def get_by_inbound_protocol(
    aws_client: AWSClient, region_name: str, protocol: str, vpc_id: Optional[str] = None
) -> List[EC2SecurityGroup]:
    """Get all security groups that have inbound rules for the supplied protocol.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __protocol:__ the inbound rule protocol (`tcp`, `udp`, or `-1` for any)

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"
    A list of [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.get_by_inbound_protocol:
        region_name: eu-west-2
        protocol: tcp
      load:
        aws_client: aws_secrets
      save: tcp_groups
    ```
    """
    filters = [{"Name": "ip-permission.protocol", "Values": [protocol]}]
    if vpc_id:
        filters.append({"Name": "vpc-id", "Values": [vpc_id]})
    return creator(
        paging_request(
            ec2_client(aws_client=aws_client, region_name=region_name).describe_security_groups, data_key, Filters=filters
        )
    )


def get_no_instance(aws_client: AWSClient, region_name: str, vpc_id: Optional[str] = None) -> List[EC2SecurityGroup]:
    """Get all security groups that are __not__ associated with an instance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"
    A list of [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.get_no_instance:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: groups_with_no_instance
    ```
    """
    return [
        group
        for group in get_all(aws_client=aws_client, region_name=region_name, vpc_id=vpc_id)
        if group["group_id"]
        not in [
            group["group_id"]
            for group in sum(
                [
                    instance["security_groups"]
                    for instance in instances.get_all(aws_client=aws_client, region_name=region_name, vpc_id=vpc_id)
                    if instance["security_groups"]
                ],
                [],
            )
        ]
    ]


def get_no_network_interface(aws_client: AWSClient, region_name: str, vpc_id: Optional[str] = None) -> List[EC2SecurityGroup]:
    """Get all security groups that are __not__ associated with a network interface.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __vpc_id__: the VPC ID to look in (defaults to _all_ VPCs for the supplied region)

??? success "Output"
    A list of [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.get_no_network_interface:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: groups_with_no_interface
    ```
    """
    return [
        group
        for group in get_all(aws_client=aws_client, region_name=region_name, vpc_id=vpc_id)
        if group["group_id"]
        not in [
            group["group_id"]
            for group in sum(
                [
                    interface["groups"]
                    for interface in network_interfaces.get_all(aws_client=aws_client, region_name=region_name)
                    if interface["groups"]
                ],
                [],
            )
        ]
    ]


def create(aws_client: AWSClient, region_name: str, name: str, description: str, vpc_id: Optional[str] = None) -> bool:
    """Create a new security group.

!!! question "Minimum Plugin Version: 2.0.0"

!!! tip "Idempotent Action"

    This action will do nothing if a security group already exists with the same name.

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __name:__ the new security group name

    - __description:__ a description for the new security group

    - __vpc_id__: the VPC ID to create the group in (defaults to the region's default VPC)

??? success "Output"

    - `true` if the security group was created
    - `false` if a security group with the same name already exists

??? example
    ```YAML linenums="1"
    - aws.ec2.security_groups.create:
        region_name: eu-west-2
        name: LDAPS
        description: LDAPS access over port 636 for Domain Controllers
      load:
        aws_client: aws_secrets
    ```
    """
    kwargs = {
        "GroupName": name,
        "Description": description,
    }
    if vpc_id:
        kwargs["VpcId"] = vpc_id

    try:
        ec2_client(aws_client=aws_client, region_name=region_name).create_security_group(**kwargs)
        return True

    except botocore.exceptions.ClientError as e:
        if "already exists" in str(e):
            # Idempotent functions shouldn't fail if the desired object already exists
            return False
        raise e


def delete(aws_client: AWSClient, region_name: str, group_id: str) -> bool:
    """Delete a security group.

!!! question "Minimum Plugin Version: 2.0.0"

!!! tip "Idempotent Action"

    This action will do nothing if the supplied group doesn't exist.

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __group_id:__ the ID of the group to delete

??? success "Output"

    - `true` if the security group was deleted
    - `false` if it does not exist

??? example

    Finding a group with [get_by_name](#get_by_name), saving it as `group`, & deleting it:

    ```YAML linenums="1" hl_lines="6 12"
    - aws.ec2.security_groups.get_by_name:
        region_name: eu-west-2
        name: Redis Access (retired)
      load:
        aws_client: aws_secrets
      save: group

    - aws.ec2.security_groups.delete:
        region_name: eu-west-2
        group_id:
      load:
        group_id: group.group_id
        aws_client: aws_secrets
    ```
    """
    try:
        ec2_client(aws_client=aws_client, region_name=region_name).delete_security_group(GroupId=group_id)
        return True
    except botocore.exceptions.ClientError as e:
        if "invalid id" in str(e).lower():
            # Idempotent functions shouldn't fail if the desired object is already deleted
            return False
        raise e


def create_inbound_rule(
    aws_client: AWSClient,
    region_name: str,
    group_id: str,
    protocol: Union[str, int],
    from_port: int,
    to_port: int,
    cidr: str,
    rule_description: Optional[str] = None,
) -> bool:
    """Create an inbound rule in a security group.

!!! question "Minimum Plugin Version: 2.0.0"

!!! tip "Idempotent Action"

    This action will do nothing if a matching inbound rule already exists in the security group.

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

    - __group_id:__ the ID of the security group

    - __protocol:__ the inbound rule protocol (`tcp`, `udp`, or `-1` for any)

    - __from_port:__ the start of the rule port range

    - __to_port:__ the end of the rule port range

    - __cidr:__ the CIDR address for permitted traffic

    - __rule_description:__ a description for the rule (defaults to empty)

??? success "Output"

    - `true` if the inbound rule was created
    - `false` if a matching inbound rule already exists

??? example

    ```YAML linenums="1"

    - aws.ec2.security_groups.create_inbound_rule:
        region_name: eu-west-2
        group_id: sg-04cd36d28fe84e079
        protocol: tcp
        from_port: 22
        to_port: 22
        cidr: 1.2.3.4/5
        rule_description: SSH Access from Corporate Office
      load:
        aws_client: aws_secrets
    ```
    """
    description_pattern = r"[a-zA-Z0-9.\s_\-\:/\(\)#,@\[\]+=&\;\{\}\!\$\*]+"
    if not re.match(re.compile(description_pattern), rule_description):
        raise exceptions.InvalidRule(f"Rule description must match the following expression: '{description_pattern}'")

    protocol = str(protocol.lower())

    if protocol not in {"udp", "tcp", "-1"}:
        raise exceptions.InvalidRule(f"Protocol must be udp, tcp, or -1. Received '{protocol}'.")

    try:
        ec2_client(aws_client=aws_client, region_name=region_name).authorize_security_group_ingress(
            GroupId=group_id,
            IpPermissions=[
                {
                    "FromPort": from_port,
                    "IpProtocol": protocol,
                    "IpRanges": [{"CidrIp": cidr, "Description": rule_description}],
                    "ToPort": to_port,
                },
            ],
        )
        return True
    except botocore.exceptions.ClientError as e:
        if "already exists" in str(e):
            # Idempotent functions shouldn't fail if the desired object already exists
            return False
        raise e


def output_table(text: str, groups: Union[EC2SecurityGroup, List[EC2SecurityGroup]]) -> None:
    """Display a list of security groups in a table.

The table will have the following columns:

- Name
- ID
- Description
- Inbound Rule Count
- Outbound Rule Count
- VPC ID

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a single or list of [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example "Single Group"

    Finding a security group with [get_by_name](#get_by_name), saving it as `group`, & displaying it:

    {% raw %}
    ```YAML linenums="1" hl_lines="6 11"
    - aws.ec2.security_groups.get_by_name:
        region_name: eu-west-2
        name: SSH Access
      load:
        aws_client: aws_secrets
      save: group

    - aws.ec2.security_groups.output_table:
        text: {{ group.name }}
      load:
        groups: group
    ```
    {% endraw %}

??? example "Multiple Groups"

    Finding all security groups with [get_all](#get_all), saving them as `groups`, & displaying them:

    ```YAML linenums="1" hl_lines="5 10"
    - aws.ec2.security_groups.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: groups

    - aws.ec2.security_groups.output_table:
        text: All Security Groups
      load:
        groups: groups
    ```
    """
    if not isinstance(groups, list):
        groups = [groups]

    ui.output_table(text, header=GROUP_TABLE_HEADER, data=[[get(group) for get in GROUP_TABLE_FIELDS] for group in groups])


def output_inbound_rule_table(text: str, group: EC2SecurityGroup) -> None:
    """Display the inbound rules for a security groups in a table.

The table will have the following columns:

- Protocol
- From Port
- To Port
- Allowed IPv4 CIDR
- Description

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __group:__ a single [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionary

??? success "Output"

    Nothing is outputted by this action.

??? example

    Finding a security group with [get_by_name](#get_by_name), saving it as `group`, & displaying the inbound rules:

    {% raw %}
    ```YAML linenums="1" hl_lines="6 11"
    - aws.ec2.security_groups.get_by_name:
        region_name: eu-west-2
        name: SSH
      load:
        aws_client: aws_secrets
      save: group

    - aws.ec2.security_groups.output_inbound_rule_table:
        text: Inbound Rules for {{ group.name }}
      load:
        group: group
    ```
    {% endraw %}
    """
    header = ["Protocol", "From Port", "To Port", "Allowed IPv4 CIDR", "Description"]
    rows = []
    for rule_set in group["inbound_rules"]:
        for ip_range in rule_set["ip_ranges"]:
            rows.append(
                [
                    rule_set["protocol"],
                    rule_set["from_port"],
                    rule_set["to_port"],
                    ip_range["cidr_ip"],
                    ip_range["description"] or "",
                ]
            )

    ui.output_table(text, header=header, data=rows)


def input_table(
    text: str, groups: List[EC2SecurityGroup], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[EC2SecurityGroup]:
    """Display a list of security groups in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- ID
- Description
- Inbound Rule Count
- Outbound Rule Count
- VPC ID

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a list of [EC2SecurityGroup](../inputs_outputs/#user) dictionaries to display in the table

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [EC2SecurityGroup](../inputs_outputs/#ec2securitygroup) dictionaries.

??? tip

    - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

    - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

??? example

    Finding security groups with [get_all](#get_all), saving them as `groups`, & waiting for a single selection:

    ```YAML linenums="1" hl_lines="5 12 13"
    - aws.ec2.security_groups.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: groups

    - aws.ec2.security_groups.input_table:
        text: Choose a security group
        minimum: 1
        maximum: 1
      load:
        groups: groups
      save: selected_group
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=GROUP_TABLE_HEADER,
            data=[[get(group) for get in GROUP_TABLE_FIELDS] for group in groups],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [groups[i] for i in indices]
