"""EC2 - Virtual Private Clouds

This module contains actions for viewing & managing EC2 VPCs.
"""
import uuid
import logging
import functools

from typing import List, Union, Optional

from ppa import ui

from ._ec2 import create_records, paging_request
from .records import EC2VPC, EC2VPCData, get_tag_value
from ..records import AWSClient
from .._api import client


VPC_TABLE_HEADER = ["Name", "ID", "CIDR", "State", "Default"]
VPC_TABLE_FIELDS = [
    lambda vpc: get_tag_value(vpc, "Name"),
    lambda vpc: vpc.get("vpc_id"),
    lambda vpc: vpc.get("cidr_block"),
    lambda vpc: vpc.get("state") or "",
    lambda vpc: str(vpc.get("is_default")),
]


logger = logging.getLogger(__name__)

ec2_client = functools.partial(client, endpoint="ec2")
creator = functools.partial(create_records, EC2VPCData)
data_key = "Vpcs"


def get_all(aws_client: AWSClient, region_name: str) -> List[EC2VPC]:
    """Get all VPCs.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __region_name:__ the EC2 region name

??? success "Output"

    A list of [EC2VPC](../inputs_outputs/#ec2vpc) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.ec2.vpcs.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: all_vpcs
    ```
    """
    return creator(paging_request(ec2_client(aws_client=aws_client, region_name=region_name).describe_vpcs, data_key))


def output_table(text: str, vpcs: Union[EC2VPC, List[EC2VPC]]) -> None:
    """Display a list of VPCs in a table.

The table will have the following columns:

- Name
- ID
- CIDR
- State
- Default

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __vpcs:__ a single or list of [EC2VPC](../inputs_outputs/#ec2vpc) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example "Multiple VPCs"

    Finding all VPCs with [get_all](#get_all), saving them as `vpcs`, & displaying them:

    ```YAML linenums="1" hl_lines="5 10"
    - aws.ec2.vpcs.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: vpcs

    - aws.ec2.vpcs.output_table:
        text: All VPCs
      load:
        vpcs: vpcs
    ```
    """
    if not isinstance(vpcs, list):
        vpcs = [vpcs]

    ui.output_table(text, header=VPC_TABLE_HEADER, data=[[get(vpc) for get in VPC_TABLE_FIELDS] for vpc in vpcs])


def input_table(text: str, vpcs: List[EC2VPC], minimum: Optional[int] = None, maximum: Optional[int] = None) -> List[EC2VPC]:
    """Display a list of VPCs in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- ID
- CIDR
- State
- Default

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __instances:__ a list of [EC2VPC](../inputs_outputs/#ec2vpc) dictionaries to display in the table

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [EC2VPC](../inputs_outputs/#ec2vpc) dictionaries.

??? tip

    - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

    - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

??? example

    Finding VPCs with [get_all](#get_all), saving them as `vpcs`, & waiting for a single selection:

    ```YAML linenums="1" hl_lines="5 12 13"
    - aws.ec2.vpcs.get_all:
        region_name: eu-west-2
      load:
        aws_client: aws_secrets
      save: vpcs

    - aws.ec2.vpcs.input_table:
        text: Choose a VPC
        minimum: 1
        maximum: 1
      load:
        vpcs: vpcs
      save: selected_vpc
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=VPC_TABLE_HEADER,
            data=[[get(vpc) for get in VPC_TABLE_FIELDS] for vpc in vpcs],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [vpcs[i] for i in indices]
