"""IAM - Groups

This module contains actions for viewing & managing IAM groups.
"""
import logging
import functools
import uuid

from typing import List, Union, Optional

from ppa import ui

from .records import IAMGroup, IAMGroupData, IAMManagedPolicy, IAMManagedPolicyData
from ._iam import paging_request, create_records
from ..records import AWSClient
from .._api import client


GROUP_TABLE_HEADER = ["Name", "ID", "Resource Name"]
GROUP_TABLE_FIELDS = ["name", "group_id", "arn"]


logger = logging.getLogger(__name__)

iam_client = functools.partial(client, endpoint="iam")
group_creator = functools.partial(create_records, IAMGroupData)
managed_policy_creator = functools.partial(create_records, IAMManagedPolicyData)
data_key = "Groups"


def get_all(aws_client: AWSClient) -> List[IAMGroup]:
    """Get all groups.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

??? success "Output"

    A list of [IAMGroup](../inputs_outputs/#iamgroup) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.iam.groups.get_all:
      load:
        aws_client: aws_secrets
      save: all_groups
    ```
    """
    logger.info("getting all IAM groups")
    return group_creator(paging_request(iam_client(aws_client=aws_client).list_groups, data_key))


def managed_policies_for_group(aws_client: AWSClient, group_name: str) -> List[IAMManagedPolicy]:
    """Get all managed policies attached to the supplied group.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __group_name:__ the name of the group

??? success "Output"

    A list of [IAMManagedPolicy](../inputs_outputs/#iammanagedpolicy) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.iam.groups.managed_policies_for_group:
      load:
        aws_client: aws_secrets
      save: managed_policies
    ```
    """
    return managed_policy_creator(
        paging_request(iam_client(aws_client=aws_client).list_attached_group_policies, "AttachedPolicies", GroupName=group_name)
    )


def output_table(text: str, groups: Union[IAMGroup, List[IAMGroup]]) -> None:
    """Display a list of groups in a table.

The table will have the following columns:

- Name
- ID
- Resource Name

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a single or list of [IAMGroup](../inputs_outputs/#iamgroup) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws.iam.groups.get_all:
      load:
        aws_client: aws_secrets
      save: all_groups

    - aws.iam.groups.output_table:
        text: All Groups
      load:
        groups: all_groups
    ```
    """
    groups = [groups] if not isinstance(groups, list) else groups
    ui.output_table(
        text, header=GROUP_TABLE_HEADER, data=[[group.get(field) or "" for field in GROUP_TABLE_FIELDS] for group in groups]
    )


def input_table(
    text: str, groups: Union[List[IAMGroup], IAMGroup], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[IAMGroup]:
    """Display a list of groups in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- ID
- Resource Name

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a single or list of [IAMGroup](../inputs_outputs/#iamgroup) dictionaries

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [Group](../inputs_outputs/#group) dictionaries.

??? tip

    - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

    - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

??? example

    Finding groups with [get_all](#get_all), saving them as `all_groups`, & waiting for a single selection:

    ```YAML linenums="1"
    - aws.iam.groups.get_all:
      load:
        aws_client: aws_secrets
      save: all_groups

    - aws.iam.groups.input_table:
        text: Choose a group
        minimum: 1
        maximum: 1
      load:
        groups: all_groups
      save: selected_group
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    groups = [groups] if not isinstance(groups, list) else groups
    while True:
        indices = ui.input_table(
            text,
            header=GROUP_TABLE_HEADER,
            data=[[group.get(field) or "" for field in GROUP_TABLE_FIELDS] for group in groups],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [groups[i] for i in indices]
