from typing import TypedDict, Optional

from datacaster.classes import CastDataClass

from .._converters import optional_date_string


DateString = Optional[str]


class IAMUserData(CastDataClass):
    name: str
    user_id: str
    created_at: Optional[DateString]
    password_last_used: Optional[DateString]
    __class_config__ = {
        "rename_fields": {
            "UserName": "name",
            "UserId": "user_id",
            "CreateDate": "created_at",
            "PasswordLastUsed": "password_last_used",
        },
        "cast_functions": {"types": {Optional[DateString]: optional_date_string}},
    }


class IAMUser(TypedDict):
    """
- Contains IAM user information

??? info "User Keys"

    __name__: the name of the user

    __user_id__: the user's ID

    __created_at__: the date & time the user was created

    __password_last_user__: the date & time the password was last used (or `null` if never used)

??? example "YAML Structure"

    ```yaml linenums="1"
    name: john.smith
    user_id: AIDAVR3TQRST3WZ2QQZTU
    created_at: 11:19:46 25/01/2021
    password_last_used: null
    ```
    """

    __annotations__ = IAMUserData.__annotations__


class IAMGroupData(CastDataClass):
    name: str
    group_id: str
    arn: str
    created_at: Optional[DateString]
    __class_config__ = {
        "rename_fields": {"GroupName": "name", "GroupId": "group_id", "Arn": "arn", "CreateDate": "created_at"},
        "cast_functions": {"types": {Optional[DateString]: optional_date_string}},
    }


class IAMGroup(TypedDict):
    """
- Contains IAM group information

??? info "User Keys"

    __name__: the name of the group

    __group_id__: the group's ID

    __arn__: the Amazon resource name

    __created_at__: the date & time the group was created

??? example "YAML Structure"

    ```yaml linenums="1"
    name: EC2Admins
    group_id: AGPAVR3TQRSTWXAIUDVV7
    arn: arn:aws:iam::[account-id]:group/EC2Admins
    created_at: 10:06:43 19/01/2021
    ```
    """

    __annotations__ = IAMGroupData.__annotations__


class IAMManagedPolicyData(CastDataClass):
    name: str
    arn: str
    __class_config__ = {
        "rename_fields": {"PolicyName": "name", "PolicyArn": "arn"},
    }


class IAMManagedPolicy(TypedDict):
    __annotations__ = IAMManagedPolicyData.__annotations__
