"""IAM - Users

This module contains actions for viewing & managing IAM user accounts.
"""
import logging
import functools
import uuid
import botocore

from typing import List, Union, Optional, Dict, Any, Callable

from ppa import ui

from . import exceptions
from .records import IAMUser, IAMUserData, IAMGroup, IAMGroupData
from ._iam import paging_request, create_records
from ..records import AWSClient
from ..exceptions import ObjectAlreadyExists, ObjectNotFound
from .._api import client


USER_TABLE_HEADER = ["Name", "ID", "Created At"]
USER_TABLE_FIELDS = ["name", "user_id", "created_at"]


logger = logging.getLogger(__name__)

iam_client = functools.partial(client, endpoint="iam")
user_creator = functools.partial(create_records, IAMUserData)
group_creator = functools.partial(create_records, IAMGroupData)
data_key = "Users"


def get_all(aws_client: AWSClient) -> List[IAMUser]:
    """Get all users.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

??? success "Output"

    A list of [IAMUser](../inputs_outputs/#iamuser) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.iam.users.get_all:
      load:
        aws_client: aws_secrets
      save: all_users
    ```
    """
    logger.info("getting all IAM users")
    return user_creator(paging_request(iam_client(aws_client=aws_client).list_users, data_key))


def get_by_name(aws_client: AWSClient, user_name: str) -> IAMUser:
    """Get the IAM user with the supplied name.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __user_name:__ the name of the user

??? success "Output"

    A single [IAMUser](../inputs_outputs/#iamuser) dictionary.

??? example
    ```YAML linenums="1"
    - aws.iam.users.get_by_name:
        user_name: john.smith
      load:
        aws_client: aws_secrets
      save: user
    ```
    """
    logger.info(f"getting IAM user with name {user_name}")
    try:
        return user_creator([iam_client(aws_client=aws_client).get_user(UserName=user_name).get("User")])
    except botocore.exceptions.ClientError as e:
        if "cannot be found" in str(e).lower():
            message = f"No user was found with name {user_name}."
            ui.output_error(message)
            raise ObjectNotFound(message)
        raise e


def exists(aws_client: AWSClient, user_name: str) -> bool:
    """Check if a user with the supplied name exists.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __user_name:__ the name of the user

??? success "Output"

    Outputs `true` if the user exists, & `false` if not.

??? example
    ```YAML linenums="1"
    - aws.iam.users.exists:
        user_name: john.smith
      load:
        aws_client: aws_secrets
      save: user_exists
    ```
    """
    logger.info(f"checking whether IAM user with name {user_name} exists")
    try:
        iam_client(aws_client=aws_client).get_user(UserName=user_name)
        return True
    except botocore.exceptions.ClientError as e:
        if "cannot be found" in str(e).lower():
            return False
        raise e


def groups_for_user(aws_client: AWSClient, user_name: str) -> List[IAMGroup]:
    """Get the groups the supplied user is in.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __user_name:__ the name of the user

??? success "Output"

    A list of [IAMGroup](../inputs_outputs/#iamgroup) dictionaries.

??? example
    ```YAML linenums="1"
    - aws.iam.users.groups_for_user:
        user_name: john.smith
      load:
        aws_client: aws_secrets
      save: user_groups
    ```
    """
    logger.info(f"getting all groups for IAM user {user_name}")
    return group_creator(paging_request(iam_client(aws_client=aws_client).list_groups_for_user, "Groups", UserName=user_name))


def create(aws_client: AWSClient, user_name: str, path: Optional[str] = None, tags: Optional[Dict[str, Any]] = None,) -> IAMUser:
    """Create a new IAM user.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __user_name:__ the name of the user

    - __path:__ an optional path for the user (defaults to `/`)

    - __tags:__ an optional dictionary of tags for the user (defaults to empty)


??? success "Output"

    A single [IAMUser](../inputs_outputs/#iamuser) dictionary.

??? example
    ```YAML linenums="1"
    - aws.iam.users.create:
        user_name: john.smith
        tags:
          team: engineering
          office: london
      load:
        aws_client: aws_secrets
      save: new_user
    ```
    """
    logger.info(f"creating user {user_name} with path {path or '/'} & {f'tags {tags}' or 'no tags'}")
    kwargs = {"UserName": user_name}
    # The boto3 client doesn't handle None values, so only add them to the args if they are not None.
    if path is not None:
        kwargs["Path"] = path
    if tags:
        kwargs["Tags"] = [{"Key": name, "Value": value} for name, value in tags.items()]

    try:
        return IAMUserData(**iam_client(aws_client=aws_client).create_user(**kwargs)["User"]).__dict__
    except botocore.exceptions.ClientError as e:
        if "already exists" in str(e).lower():
            message = "Failed to create the user as another user already exists with the same name."
            ui.output_error(message)
            raise ObjectAlreadyExists(message)
        ui.output_error(f"Failed to create user: {str(e)}")
        raise e


def delete(aws_client: AWSClient, user_name: str) -> None:
    """Delete an IAM user.

This action also deletes the user's:

- Access keys
- Signing certificate
- SSH public key
- Git credentials
- Multi-factor authentication (MFA) device
- Inline policies
- Attached managed policies

!!! question "Minimum Plugin Version: 3.0.0"

!!! tip "Virtual MFA Devices"

    This action will not deactivate any virtual MFA devices assigned to the user.

    If the user has any virtual MFA devices assigned to them, this action may fail.

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __user_name:__ the name of the user

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - aws.iam.users.delete:
        user_name: john.smith
      load:
        aws_client: aws_secrets
    ```
    """

    def _delete_refs(*, get_func: Callable, delete_func: Callable, data_key: str, identifier: str) -> None:
        for value in paging_request(get_func, data_key, UserName=user_name):
            if isinstance(value, dict):
                # Most get functions return a list of dictionaries...
                delete_func(UserName=user_name, **{identifier: value[identifier]})
            elif isinstance(value, list):
                # ...but some just return a list of string values
                delete_func(UserName=user_name, **{identifier: value})
            else:
                raise ValueError(f"Expected a list or dictionary from the boto3 function. Received: {value}.")

    def _error_tip():
        ui.output_error("Deleting the user failed. There may be objects still attached to the user preventing deletion.")
        ui.output_error("Please view the logs for more information.")

    _iam_client = iam_client(aws_client=aws_client)

    try:
        logger.info(f"deleting login profile for user {user_name}")
        _iam_client.delete_login_profile(UserName=user_name)

        logger.info(f"deleting any access keys for user {user_name}")
        _delete_refs(
            get_func=_iam_client.list_access_keys,
            delete_func=_iam_client.delete_access_key,
            data_key="AccessKeyMetadata",
            identifier="AccessKeyId",
        )

        logger.info(f"deleting any signing certificates for user {user_name}")
        _delete_refs(
            get_func=_iam_client.list_signing_certificates,
            delete_func=_iam_client.delete_signing_certificate,
            data_key="Certificates",
            identifier="CertificateId",
        )

        logger.info(f"deleting any ssh public keys for user {user_name}")
        _delete_refs(
            get_func=_iam_client.list_ssh_public_keys,
            delete_func=_iam_client.delete_ssh_public_key,
            data_key="SSHPublicKeys",
            identifier="SSHPublicKeyId",
        )

        logger.info(f"deleting any service specific credentials for user {user_name}")
        _delete_refs(
            get_func=_iam_client.list_service_specific_credentials,
            delete_func=_iam_client.delete_service_specific_credential,
            data_key="ServiceSpecificCredentials",
            identifier="ServiceSpecificCredentialId",
        )

        logger.info(f"deactivating any mfa devices for user {user_name}")
        _delete_refs(
            get_func=_iam_client.list_mfa_devices,
            delete_func=_iam_client.deactivate_mfa_device,
            data_key="MFADevices",
            identifier="SerialNumber",
        )

        logger.info(f"deleting any inline policies for user {user_name}")
        _delete_refs(
            get_func=_iam_client.list_user_policies,
            delete_func=_iam_client.delete_user_policy,
            data_key="PolicyNames",
            identifier="PolicyName",
        )

        logger.info(f"removing any groups for user {user_name}")
        _delete_refs(
            get_func=_iam_client.list_groups_for_user,
            delete_func=_iam_client.remove_user_from_group,
            data_key="Groups",
            identifier="GroupName",
        )

        logger.info(f"deleting user {user_name}")
        _iam_client.delete_user(UserName=user_name)
    except botocore.exceptions.ClientError as e:
        if "cannot be found" in str(e).lower():
            message = f"No user was found with name {user_name}."
            ui.output_error(message)
            raise ObjectNotFound(message)
        _error_tip()
    except Exception as e:
        _error_tip()
        raise e


def set_password(aws_client: AWSClient, user_name: str, password: str, reset_required: Optional[bool] = False) -> None:
    """Set the password for an IAM user.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __user_name:__ the name of the user

    - __password:__ the password to set

    - __reset_required:__ set to `true` to force a password change when the user logs on (defaults to `false`)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa.ui.input_password:
        text: New password
      save: new_password

    - aws.iam.users.set_password:
        user_name: john.smith
        reset_required: true
      load:
        password: new_password
        aws_client: aws_secrets
    ```
    """
    try:
        iam_client(aws_client=aws_client).create_login_profile(
            UserName=user_name, Password=password, PasswordResetRequired=reset_required
        )
    except botocore.exceptions.ClientError as e:
        if any([err in str(e).lower() for err in {"password should meet", "should have a minimum length"}]):
            message = "Failed to set the new password as it does not meet the required complexity and/or length."
            ui.output_error(message)
            raise exceptions.PasswordComplexityError(message)
        raise e


def add_to_group(aws_client: AWSClient, user_name: str, group_name: str) -> None:
    """Add an IAM user to a group.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __user_name:__ the name of the user

    - __group_name:__ the name of the group

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - aws.iam.users.add_to_group:
        user_name: john.smith
        group_name: ec2_auditors
      load:
        aws_client: aws_secrets
    ```
    """
    logger.info(f"adding user {user_name} to group {group_name}")
    return iam_client(aws_client=aws_client).add_user_to_group(UserName=user_name, GroupName=group_name)


def remove_from_group(aws_client: AWSClient, user_name: str, group_name: str) -> None:
    """Remove an IAM user from a group.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __aws_client:__ an [AWSClient](../inputs_outputs/#awsclient) dictionary

    - __user_name:__ the name of the user

    - __group_name:__ the name of the group

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - aws.iam.users.remove_from_group:
        user_name: john.smith
        group_name: ec2_auditors
      load:
        aws_client: aws_secrets
    ```
    """
    logger.info(f"removing user {user_name} from group {group_name}")
    return iam_client(aws_client=aws_client).remove_user_from_group(UserName=user_name, GroupName=group_name)


def output_table(text: str, users: Union[IAMUser, List[IAMUser]]) -> None:
    """Display a list of users in a table.

The table will have the following columns:

- Name
- ID
- Created At

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ a single or list of [IAMUser](../inputs_outputs/#iamuser) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws.iam.users.get_all:
      load:
        aws_client: aws_secrets
      save: all_users

    - aws.iam.users.output_table:
        text: All Users
      load:
        users: all_users
    ```
    """
    users = [users] if not isinstance(users, list) else users
    ui.output_table(
        text, header=USER_TABLE_HEADER, data=[[user.get(field) or "" for field in USER_TABLE_FIELDS] for user in users]
    )


def input_table(
    text: str, users: Union[List[IAMUser], IAMUser], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[IAMUser]:
    """Display a list of users in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- ID
- Created At

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ a single or list of [IAMUser](../inputs_outputs/#iamuser) dictionaries

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [User](../inputs_outputs/#user) dictionaries.

??? tip

    - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

    - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

??? example

    Finding users with [get_all](#get_all), saving them as `all_users`, & waiting for a single selection:

    ```YAML linenums="1"
    - aws.iam.users.get_all:
      load:
        aws_client: aws_secrets
      save: all_users

    - aws.iam.users.input_table:
        text: Choose a user
        minimum: 1
        maximum: 1
      load:
        users: all_users
      save: selected_user
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    users = [users] if not isinstance(users, list) else users
    while True:
        indices = ui.input_table(
            text,
            header=USER_TABLE_HEADER,
            data=[[user.get(field) or "" for field in USER_TABLE_FIELDS] for user in users],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [users[i] for i in indices]
